import re
import xml.etree


class ParseCommand:
    def parse(self, marc_xml: xml.etree) -> str:
        raise NotImplementedError


class SelectorPredicate:
    pat_sel = re.compile(r"\$\$(\d) ([a-z]{3})")

    def __init__(self, selector):
        # Debug Output
        # print(f'  Initializing SelectorPredicate for selector "{selector}"')
        match_result = self.pat_sel.match(selector)
        if not match_result:
            raise ValueError(f'String "{selector}" does not match pattern {self.pat_sel.pattern}')
        self.subfield = match_result.group(1)
        self.value = match_result.group(2)

    def __call__(self, element):
        for child in element:
            if child.attrib.get('code', '') == self.subfield:
                if child.text.strip() == self.value:
                    return True
        return False


class SingleCommand(ParseCommand):
    # Testing the input of the patterns, we can either have:
    # * tag only (if it is a controlfield, like 009)
    # * tag with inds only
    # * tag with both inds and subfield
    # Indicators must always be given in pairs, blank-separated
    pat_tag = re.compile(r"^(\d{3}|AVA)( ([0-9_*]) ([0-9_*]))?"
                         r"( \$\$[a-zA-Z0-9])?$")

    def __init__(self, pattern, selector=None, join_char=None):
        # Debug Output
        # print(f'Initializing SingleCommand for pattern "{pattern}"')
        if selector:
            # print(f'   with selector {selector}')
            self.selector = SelectorPredicate(selector)
        else:
            self.selector = None
        if join_char:
            # print(f'   collecting all results separated by "{join_char}"')
            pass

        self.join_char = join_char

        match_result = self.pat_tag.match(pattern)

        if not match_result:
            raise ValueError(f'String "{pattern}" does not match pattern {self.pat_tag.pattern}')

        tag = match_result.group(1)
        # ind1 is group 3, since group 2 is both of the inds together
        ind1 = match_result.group(3)
        ind2 = match_result.group(4)
        subfield = match_result.group(5)

        d = {'tag': tag}

        def handle_ind(s, key):
            if s is None or s == '_':
                d[key] = ' '
            elif s == '*' or tag.startswith('00'):
                # Control fields (starting with '00') don't have 'indX' attributes
                # Omitting attributes _should_ effect a wildcard behaviour (think '*')
                pass
            else:
                d[key] = s

        handle_ind(ind1, 'ind1')
        handle_ind(ind2, 'ind2')
        self.attrib = d

        if subfield is not None:
            self.subfield = subfield.strip(' $')

        # only controlfields (00X) may have no subfield specified
        if (not tag.startswith('00') and
                subfield is None):
            raise ValueError('No subfield specified - only controlfields (00X) may have no subfield specified')

        if tag.startswith('00'):
            del(self.attrib['ind1'])
            del(self.attrib['ind2'])

    def parse(self, marc_xml: xml.etree):
        collector = []
        for element in marc_xml:
            if self.attrib.items() <= element.attrib.items():
                if self.selector:
                    if not self.selector(element):
                        continue
                # special handling of control fields (00X, no subfields)
                if self.attrib['tag'].startswith('00'):
                    collector.append(element.text)
                # currently there is no default for the subfield
                for subfield in element:
                    if subfield.attrib == {'code': self.subfield}:
                        collector.append(subfield.text)
        if not collector:
            return ''
        elif not self.join_char:
            return collector[0]
        else:
            return self.join_char.join(collector)


class PreferenceCommand(ParseCommand):
    def __init__(self, pattern, selector=None, join_char=None):
        patterns = [pat.strip() for pat in pattern.split('>')]
        if selector:
            selectors = [sel.strip() for sel in selector.split('>')]
            self.commands = [SingleCommand(pat, sel, join_char) for pat, sel in zip(patterns, selectors)]
        else:
            self.commands = [SingleCommand(pat, join_char=join_char) for pat in patterns]

    def parse(self, marc_xml: xml.etree) -> str:
        for com in self.commands:
            value = com.parse(marc_xml)
            if value != '':
                return value
        return ''


class ConcatCommand(ParseCommand):
    def __init__(self, pattern, selector=None, join_char=None):
        patterns = [pat.strip() for pat in pattern.split(';')]
        if selector:
            selectors = [sel.strip() for sel in selector.split(';')]
            self.commands = [PreferenceCommand(pat, sel, join_char) for pat, sel in zip(patterns, selectors)]
        else:
            self.commands = [PreferenceCommand(pat, join_char=join_char) for pat in patterns]

    def parse(self, marc_xml: xml.etree) -> str:
        values = [com.parse(marc_xml) for com in self.commands]
        non_empty_values = [part for part in values if part != '']
        return '; '.join(non_empty_values) if non_empty_values else ''


def build_extractor(pattern, selector: str = None, collector_character: str = None) -> ParseCommand:
    """
    Create a parser object for extracting specific fields or subfields
    from a MARC-XML etree.

    The pattern given has to conform to the following grammar (EBNF):

    pattern = preferencePattern, { ";", preferencePattern }
    preferencePattern = singlePattern, { ">", singlePattern }
    singlePattern = tag, [ ind1, [ ind2, [ field ]]]
    tag = DIGIT, DIGIT, DIGIT
    ind1 = DIGIT
    ind2 = DIGIT
    field = "$$", ( DIGIT | LOWER_ALPHA )

    The selector given has to conform to the following grammar (EBNF):

    selector = preferenceSelector, { ";", preferenceSelector }
    preferenceSelector = singleSelector { ">", singleSelector }
    singleSelector = field, content
    field = "$$", ( DIGIT | LOWER_ALPHA )
    content = LOWER_ALPHA, LOWER_ALPHA, LOWER_ALPHA


    :param pattern:
    :param selector:
    :param collector_character: Set if you expect the pattern to match more than once.
    Use the character given to separate the collected values in the output.
    :return:
    """
    root_command = ConcatCommand(pattern, selector, collector_character)
    return root_command
