"""almasru with ONB defaults."""

import re

from . import almasru


class OnbRecordRetriever(almasru.RecordRetriever):

    def __init__(self,
                 subdomain: str = 'obv-at-oenb',
                 inst_code: str = '43ACC_ONB',
                 xml_format: str = 'marcxml'):
        super().__init__(subdomain, inst_code, xml_format)

        self.namespaces = {
            'srw': 'http://www.loc.gov/zing/srw/',
            'diag': 'http://www.loc.gov/zing/srw/diagnostic',
            'marc': 'http://www.loc.gov/MARC21/slim'
        }

        self.dtl_re = re.compile(r'dtl\/\d*')
        self.pid_re = re.compile(r'\d+')

    # Create alias for querying AC-numbers
    def by_ac_num(self, ac_num: str):
        return self.by_marc_009(ac_num)

    def get_pids(self, xml) -> list:
        """
        Extract all PIDs from XML based on MARC 856 4 2.
        PIDs are identifiers migrated from a legacy system.

        :param xml: MARC xml as xml.etree.ElementTree Element
        :return: List of identifiers.
        """

        datafields = xml.findall("marc:datafield", self.namespaces)

        pids = []

        for datafield in datafields:
            if datafield.attrib == {'tag': '856', 'ind1': '4', 'ind2': '2'}:

                marc_856_url = self._get_subfield(datafield, 'u').text
                marc_856_note = self._get_subfield(datafield, 'x').text

                if 'ONB-DTL' in marc_856_note:
                    dtl_suffix = self.dtl_re.findall(marc_856_url)[0]
                    pids.append(self.pid_re.findall(dtl_suffix)[-1])

        return pids

    def _get_subfield(self, element, subfield):
        xpath = "marc:subfield[@code='"+subfield+"']"
        return element.find(xpath, self.namespaces)
